#ifndef __CONTROL__HPP
#define __CONTROL__HPP

#include <fstream>
#include "real3d.hpp"
#include "setmpi.hpp"
#include "command.hpp"

namespace dpd{


class Control{
protected:
    Command* command;
    SetMPI *mpi;

    std::string control_fname;
    std::ifstream constream;

    /* Control variables
     * If one wants to add varialbes to read from the
     * control file, add variables here 
     * read it from Control::readControl(). They
     * have to be broadcasted in Control:bcastControl()
     * If necessary, set the default values in 
     * Control::setDefaults().
     */
    bool restart;
    real temperature;
    real dt;
    real gamma;
    real lambda;
    real cellcut;
    real saferatio;
    int randomseed;
    int integrator;
    int totalsteps;

    bool dumpbinary;
    int trajfreq;
    int stressfreq;
    int logfreq;
    int forcefreq;

    int nonbonded;
    int bondlength;

    int walltype;
    Ivec walldirect;
    Rvec wallmin;
    Rvec wallmax;

    int rmcomvfreq;
    Ivec rmcomvdir;

    bool dumpfrozen;

    real targetforce;

    real eminmaxforce;
   
    /*Control variables for external fields*/
    real pull_springk;
    Real3D pull_coord;
    int pull_direct;

    /*Control variables for gravitation field*/
    real gravity_field;
    int gravity_direct;


    /*Control varialbes for box deformation*/
    int deformation;
    int deformfreq;
    real deformfactor;

    /*Control variables for particle pinning*/
    Ivec pindirect;
    std::string pinmolname;
    Ivec pinindex;
    bool pinrandom;
    int npinpermol;


    /*control variables for uniaxial flow*/
    Ivec uniflow;
    real flowrate;
    Rvec sheartensor;
    void buildShearTensor();

    /*control variables for uniaxial flow*/
    int sheardir;
    real shearrate;
    Svec liquidgrps;
    
    /*control variables for wall-induced shear*/
    std::string shrwallgrp1, shrwallgrp2;
    int wsheardir;
    real wshearrate;

    
    Svec tokens;
    std::string line;

    
    /*control variables for particle pulling*/
    /*
    Ivec ppulldirect;
    real pullvel;
    real pullk;
    std::string pullgroup;
    */

    /*control variables for slip spring*/
    bool slipspring;
    int numss;
    int nsteps_ssmc;
    int nsteps_ssdpd;
    real ssl0;
    real ssk;
    real sscutmax;
    real sscutmin;
    int sspoltype;
    real intrassbias;




public:
    Control(){}
    Control(Command* command, SetMPI *mpi);
    Control(std::string control_fname, SetMPI *mpi);
    ~Control(){};

    Command* getCommand(){return command;}
    void setDefaults();
    bool getValues(std::string option, int numvalues, Svec& values);
    void readControl();
    void setDependency();
    void bcastControl();

    bool getRestart(){return restart;}

    real getTemperature(){ return temperature; }
    real getTimeStep(){ return dt;}
    real getGamma(){ return gamma; }
    real getLambda(){ return lambda; }
    real getCellCutoff(){ return cellcut; }
    real getSafeRatio(){ return saferatio;}
    int getRandomSeed(){ return randomseed; }
    int getIntegrator(){ return integrator; }
    int getTotalSteps(){ return totalsteps; }
    int getTrajFrequency(){ return trajfreq; }
    int getLogFrequency(){ return logfreq; }
    int getStressFrequency(){ return stressfreq; }
    int getForceFrequency(){ return forcefreq; }

    int getNonbondedInteraction(){ return nonbonded; }
    int getBondLengthInteraction(){ return bondlength; }
    real getTargetForce(){ return targetforce; }

    bool getDumpFrozen() { return dumpfrozen; }
    bool getDumpBinary(){ return dumpbinary; }

    real getPullingSpringK(){ return pull_springk; }
    Real3D getPullingCoord(){ return pull_coord; }
    int getPullingDirect(){ return pull_direct; }

    real getGravityField(){ return gravity_field; }
    int getGravityDirection(){ return gravity_direct; }

    int getWallType(){ return walltype; }
    Ivec getWallDirection(){ return walldirect; }
    Rvec getWallMinPosition(){ return wallmin; }
    Rvec getWallMaxPosition(){ return wallmax; }

    int getRmCOMVelFrequency(){ return rmcomvfreq; }
    Ivec getRmCOMVelDirection(){ return rmcomvdir; }


    int getDeformation(){ return deformation; }
    int getDeformationFrequency(){ return deformfreq; }
    real getDeformationFactor(){ return deformfactor; }

    bool getPinning();
    Ivec& getPinningDirection(){ return pindirect; }
    std::string& getPinningMoleculeName(){ return pinmolname; }
    Ivec& getPinningIndex(){ return pinindex; }
    bool& getPinningRandom(){ return pinrandom; }
    int& getNumPinningPerMolecule(){ return npinpermol; }

    bool isUniflow();
    Rvec getShearTensor(){ return sheartensor; }

    bool isWallSheared();
    int getWallShearDir(){ return wsheardir; }
    real getWallShearRate(){ return wshearrate; }
    std::string getShearWallGroup1(){ return shrwallgrp1; }
    std::string getShearWallGroup2(){ return shrwallgrp2; }


    bool slipSpring(){ return slipspring; }
    int getNumSlipSprings(){ return numss;}
    int getNumMCSeqSteps(){ return nsteps_ssmc; }
    int getNumDPDSeqSteps(){ return nsteps_ssdpd; }
    real getSlipSpringLength(){ return ssl0; }
    real getSlipSpringForceConst(){ return ssk; }
    real getSlipSpringMaxCutoff(){ return sscutmax; }
    real getSlipSpringMinCutoff(){ return sscutmin; }
    int getSlipSpringPolymerType(){ return sspoltype; }
    int getIntraSlipSpringBias(){ return intrassbias; }


    Svec getLiquidGroups(){ return liquidgrps; }



};
}


#endif
